#include "common.h"
#include <assert.h>

#define HASH_LIST_SIZE 101

/* Defines for attempting to share memory */
/* max bucket size for value */
#define INTMC_VALUE_SIZE (1024 * 1024)
/* num buckets, for now. Deal with later */
#define INTMC_BUCKET_NUM 128

typedef struct bucket bucket;
typedef struct internal_server_state_st internal_server_state_st;

/* debugging, wooo! */
#define DP(x)	printf("%s\n", x);

struct bucket {
    uint32_t flags;
    time_t expiration;
    /* char *value; */
    char value[INTMC_VALUE_SIZE];
    bucket *next;
    bucket *read_chain;
    size_t value_length;
    size_t value_array_size;
    size_t key_length;
    uint64_t cas;
    char key[250];
    char in_use;
};

struct internal_server_state_st {
  bucket *read_chain;
  /* bucket *list[HASH_LIST_SIZE]; */
  bucket **list;
  bucket **holding; /*  I think this can hold buckets until they get placed
			into the **list. */
  int refcount;
};


static bool is_started(memcached_st *ptr)
{
  return ptr->hosts[0].embedded_server ? true : false;
}

static internal_server_state_st *get_instance(memcached_st *ptr)
{
  internal_server_state_st *instance;

  if (ptr->hosts[0].embedded_server == NULL)
  {
    instance= ptr->call_calloc(ptr, 1, sizeof(internal_server_state_st)); 
    /* better to allocate a bit much */
    libmemcached_shm_init(ptr, ((sizeof(bucket) * INTMC_BUCKET_NUM)) + (sizeof(bucket) * HASH_LIST_SIZE));
    bucket *blist;
    bucket **listtmp;
    /* OK, I think that should have done it. Theoretically, we have a shared
     * bucket list */

    if (instance == NULL)
      return NULL;
    char *hrmph;
    hrmph = ptr->shm_area->shm;
    listtmp = (bucket **)hrmph;
    hrmph += (sizeof(bucket **) * HASH_LIST_SIZE);
    blist = (bucket *)hrmph;
    /* listtmp = (bucket **)(ptr->shm_area->shm); */
    /* blist = (bucket *)(ptr->shm_area->shm + (sizeof(char) * HASH_LIST_SIZE)); */
    /* blist = ((bucket *)ptr->shm_area->shm + (sizeof(bucket **) * HASH_LIST_SIZE)); */
    /* blist = (listtmp + (sizeof(char) * HASH_LIST_SIZE)); */
    /* printf("try to set listtmp?\n");
    (*listtmp)->in_use = 0;
    printf("able to set something with listtmp\n");
    */
    printf("Dear god, what the hell is wrong?\n");
    blist->in_use = 0;
    printf("able to set blist\n"); 
    printf("can we set blist to listtmp?\n");
    (*listtmp) = blist;
    printf("apparently so\n");
    instance->holding = &blist;
    instance->list = &(*listtmp);
    ptr->hosts[0].embedded_server= instance;
  }
  else
     instance= ptr->hosts[0].embedded_server;

  return instance;
}


static uint32_t crap_hash(const char *key, size_t key_length)
{
  uint32_t total= 0;
  const char *ptr;

  for (ptr= key; ptr != key + key_length; ptr++)
    total+= (uint32_t)abs(*ptr);

  return (total % HASH_LIST_SIZE);
}

#ifdef NOT_USED2
static void print_buckets(internal_server_state_st *instance, const char *key, size_t key_length)
{
  uint32_t x;

  if (key_length)
    printf("Looking for %.*s\n", (int)key_length, key);

  for (x= 0; x < HASH_LIST_SIZE; x++)
  {
    bucket *ptr;

    if ((instance->list + x) == NULL)
      continue;

    printf("Bucket %u ->\n", x);

    for (ptr= (instance->list + x); ptr; ptr= ptr->next)
      printf("\tKey %.*s\n", (int)ptr->key_length, ptr->key);

    printf("\n");
  }
}

void print_internal_hash(internal_server_state_st *instance)
{
  printf("\n-------INTERAL HASH-------\n");
  print_buckets(ptr, NULL, 0);
  printf("\n-------INTERAL HASH END---\n");
}

static void print_read_chain(internal_server_state_st *instance)
{
  bucket *ptr;

  printf("\n-------READ CHAIN----------\n");

  for (ptr= instance->read_chain; ptr; ptr= ptr->read_chain)
    printf("\tKey %.*s --> %.*s\n", (int)ptr->key_length, ptr->key, (int)ptr->value_length, ptr->value);
  printf("-------READ CHAIN END------\n\n");
}
#endif

static void delete_buckets(memcached_st *memc, bucket *ptr)
{
  bucket *next;

  if (ptr == NULL)
    return;
  /* sigh - shut up compiler warnings for a bit */
  if (memc == NULL)
    return;
  for (; ptr; ptr= next)
  {
    next= ptr->next;
    /* memc->call_free(memc, ptr->value);
    memc->call_free(memc, ptr);
    */
    /* We need: Semaphore to clear this. */
  }
}

static bucket *find_bucket(internal_server_state_st *instance, const char *key, size_t key_length)
{
  uint32_t hash_key= crap_hash(key, key_length);
  bucket *ptr;

  assert(hash_key < HASH_LIST_SIZE);
  DP("Trying to fetch the bucket.");
  for (ptr= (*(instance->list) + hash_key); ptr; ptr= ptr->next)
  {
    DP("Do we even survive the for loop?")
	printf("ptr is at? %p\n", (void *)ptr);
    DP("Can we *do* anything with the pointer?")
    printf("key could be: %s\n", ptr->key);
    DP("Can we set in_use?")
    ptr->in_use = 0;
    DP("Apparently so")
    if(key_length ==  ptr->key_length){
	DP("We can compare key length")
	}
    else {
	DP("Still alive after comparing")
	}
    if (memcmp(ptr->key, key, key_length) == 0)
    	DP("I doubt we'll reach this for now.")
    else
	DP("memcmp wasn't successful, but survived without bus error.")
    if (key_length ==  ptr->key_length && (memcmp(ptr->key, key, key_length) == 0)){
      DP("I'm guessing this doesn't.")
      return ptr;
	}
  }

  return NULL;
}

static bool
delete_bucket(memcached_st *memc, internal_server_state_st *instance, const char *key, size_t key_length)
{
  uint32_t hash_key= crap_hash(key, key_length);
  bucket *ptr;
  bucket *prev;

  ptr= *(instance->list + hash_key);

  if (ptr == NULL)
    return false;
  if (memc == NULL){ ; }

  if (key_length ==  ptr->key_length && memcmp(ptr->key, key, key_length) == 0)
  {
    *(instance->list + hash_key) = ptr->next;

    /* NEED: Semaphore, figure out best way to clear the buckets. */
    /* memc->call_free(memc, ptr->value);
    memc->call_free(memc, ptr);
    */

    return true;
  }

  for (prev= ptr; ptr; ptr= ptr->next)
  {
    if (key_length ==  ptr->key_length && memcmp(ptr->key, key, key_length) == 0)
    {
      prev->next= ptr->next;
      /* Same as above ^^ */
      /* memc->call_free(memc, ptr->value);
      memc->call_free(memc, ptr);
      */

      return true;
    }
    prev= ptr;
  }

  return false;
}

static bucket *
add_bucket(memcached_st *memc, internal_server_state_st *instance, const char *key, size_t key_length)
{
  uint32_t hash_key= crap_hash(key, key_length);
  bucket *new_bucket; 
  bucket *ptr;
  bucket *arrgh;

  new_bucket = NULL; /* Happy now? */
  DP("In add bucket")

  /* new_bucket= memc->call_calloc(memc, 1, sizeof(bucket)); */
  if (memc == NULL) {
	;
	}
  int i;
  for (i = 0; i < INTMC_BUCKET_NUM; i++){
 	DP("in the for loop to assign a bucket")
	arrgh = (*(instance->holding) + i);
	DP("Set arrgh")
	arrgh->in_use = 0;
	if (arrgh->in_use){
		DP("Well we saw it was in use")
		continue;
		}
	DP("Survived seeing if it was in use")
	new_bucket = (*(instance->holding) + i); 
  	DP("We seem to have been able to set new_bucket to a holding bucket")
	break;	
	}
  if (new_bucket == NULL) 
    return NULL;

  ptr= ((*instance->list) + hash_key);

  if (ptr == NULL)
  {
    *((instance->list) + hash_key) = new_bucket; 

    return new_bucket;
  }

  for (; ptr->next; ptr= ptr->next) {};

  ptr->next= new_bucket;

  return new_bucket;
}

void internal_cursor_reset(memcached_st *ptr)
{
  if (is_started(ptr))
  {
    internal_server_state_st *instance;

    instance= get_instance(ptr);

    instance->read_chain= NULL;
  }
}


memcached_return internal_mget(memcached_st *ptr,
                               const char * const *keys,
                               const size_t *key_length,
                               unsigned int number_of_keys)
{
  uint32_t x;
  bucket *prev= NULL;
  internal_server_state_st *instance;

  instance= get_instance(ptr);
  DP("Staring get")
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  DP("didn't get instance")
  internal_cursor_reset(ptr);
  DP("survived curor reset")
  for (x= 0; x < number_of_keys; x++)
  {
    bucket *bucket_ptr;
    bucket_ptr= find_bucket(instance, keys[x], key_length[x]);
    DP("I have a feeling this is it")

    if (bucket_ptr == NULL || bucket_ptr->value_length == 0)
      continue;

    if (instance->read_chain == NULL)
      prev= instance->read_chain= bucket_ptr;
    else
    {
      assert(prev);
      prev->read_chain= bucket_ptr;
      prev= bucket_ptr;
    }

    prev->read_chain= NULL; /* We do this to remove old read chains */
  }

  return MEMCACHED_SUCCESS;
}

memcached_return
internal_fetch(memcached_st *ptr, memcached_result_st *result)
{
  memcached_return rc;
  static bucket *value_bucket;
  internal_server_state_st *instance;

  instance= get_instance(ptr);
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  if (instance->read_chain == NULL)
    return MEMCACHED_END;

  value_bucket= instance->read_chain;
  instance->read_chain= instance->read_chain->read_chain;
  value_bucket->read_chain= NULL; /* Clean up after yourself */

  memcpy(result->key, value_bucket->key, value_bucket->key_length);
  result->key[value_bucket->key_length]= 0;

  result->key_length= value_bucket->key_length;
  result->flags= value_bucket->flags;
  result->cas= value_bucket->cas;
  result->expiration= value_bucket->expiration;

  rc= memcached_string_reset(&result->value);

  if (rc != MEMCACHED_SUCCESS)
    return rc;

  rc= memcached_string_append(&result->value, value_bucket->value, value_bucket->value_length);

  if (rc != MEMCACHED_SUCCESS)
    return rc;

  return rc;
}

memcached_return memcached_internal_storage(memcached_st *ptr,
                                            const char *master_key  __attribute__((unused)),
                                            size_t master_key_length __attribute__((unused)),
                                            const char *key,
                                            size_t key_length,
                                            const char *value,
                                            size_t value_length,
                                            time_t expiration,
                                            uint32_t flags,
                                            uint64_t cas,
                                            memcached_storage_action verb)
{
  /* char *new_ptr; */
  bucket *bucket_ptr;
  size_t malloc_length;
  internal_server_state_st *instance;

  memcached_return rc= memcached_validate_key_length(key_length, false);
  unlikely (rc != MEMCACHED_SUCCESS)
    return rc;
  DP("Did we successfully get here?")
  instance= get_instance(ptr);
  DP("Got the instance")
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  switch (verb)
  {
  case SET_OP: /* Find the bucket, if it does not exist, add it. */
  case PREPEND_OP:
  case APPEND_OP:
    DP("Fetching?")
    bucket_ptr= find_bucket(instance, key, key_length);
    DP("Survived looking for bucket")
    if (bucket_ptr == NULL)
      bucket_ptr= add_bucket(ptr, instance, key, key_length);
    DP("Bucket added?")
    /* If allocation fails, toss an error */
    if (bucket_ptr == NULL)
      return MEMCACHED_MEMORY_ALLOCATION_FAILURE;
    DP("Well then")
    break;
  case CAS_OP:
  case REPLACE_OP: /* Find the bucket, if it exists add it, otherwise toss an error. */
    bucket_ptr= find_bucket(instance, key, key_length);
    if (bucket_ptr == NULL)
    {
      if (ptr->flags & MEM_NOREPLY)
        return MEMCACHED_SUCCESS;
      return MEMCACHED_NOTFOUND;
    }

    break;
  case ADD_OP: /* Find the bucket, if it exists then error, otherwise add */
    bucket_ptr= find_bucket(instance, key, key_length);
    if (bucket_ptr)
    {
      if (ptr->flags & MEM_NOREPLY)
        return MEMCACHED_SUCCESS;
      return MEMCACHED_DATA_EXISTS;
    }

    bucket_ptr= add_bucket(ptr, instance, key, key_length);

    /* If allocation fails, toss an error */
    if (bucket_ptr == NULL)
      return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

    break;
  default:
    return MEMCACHED_SERVER_ERROR; /* This is impossible, fixes issue for compiler warning in VisualStudio */
  };

  if (verb == CAS_OP && bucket_ptr->cas != cas)
  {
    if (ptr->flags & MEM_NOREPLY)
      return MEMCACHED_SUCCESS;
    return MEMCACHED_DATA_EXISTS;
  }


  malloc_length= value_length;
  if (verb == PREPEND_OP || verb == APPEND_OP)
    malloc_length+= bucket_ptr->value_length;

  /* For the moment we just shrink if smaller */
  /* Actually, with the preallocated buckets I don't think we need to do this
   * at all. Let's see if I'm right. */
  /* new_ptr= ptr->call_realloc(ptr, bucket_ptr->value, malloc_length + 1); // +1 for the NULL
  if (new_ptr == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;
  bucket_ptr->value= new_ptr;
  */

  switch (verb)
  {
  case PREPEND_OP:
    memmove(bucket_ptr->value + value_length, bucket_ptr->value, bucket_ptr->value_length);
    memcpy(bucket_ptr->value, value, value_length);
    value_length= malloc_length;
    break;
  case  APPEND_OP:
    memcpy(bucket_ptr->value + bucket_ptr->value_length, value, value_length);
    value_length= malloc_length;
    break;
  case SET_OP:
  case REPLACE_OP:
  case ADD_OP:
  case CAS_OP:
      memcpy(bucket_ptr->value, value, value_length);
      break;
  default:
     /* Impossible */
     abort();
  }

  /*
    For now we add in the null byte so that we can eventually fix the code
    to allow hash values be used without the cost of malloc.
  */
  bucket_ptr->value[value_length]= 0;
  bucket_ptr->value_length= value_length;

  memcpy(bucket_ptr->key, key, key_length);
  bucket_ptr->key[key_length]= 0;

  bucket_ptr->key_length= key_length;
  bucket_ptr->expiration= expiration;
  bucket_ptr->flags= flags;
  bucket_ptr->cas++;

  return MEMCACHED_SUCCESS;
}

memcached_return internal_delete(memcached_st *ptr, const char *key, size_t key_length)
{
  bool test;
  internal_server_state_st *instance;

  if (is_started(ptr) == false)
  {
    if (ptr->flags & MEM_NOREPLY)
      return MEMCACHED_SUCCESS;
    return MEMCACHED_NOTFOUND;
  }

  instance= get_instance(ptr);
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  test= delete_bucket(ptr, instance, key, key_length);

  if (test)
    return MEMCACHED_SUCCESS;

  unlikely (ptr->flags & MEM_NOREPLY)
    return MEMCACHED_SUCCESS;

  return MEMCACHED_NOTFOUND;
}

memcached_return memcached_internal_flush(memcached_st *ptr, time_t expiration)
{
  uint32_t x;
  (void)ptr;
  (void)expiration;
  internal_server_state_st *instance;

  if (is_started(ptr) == false)
    return MEMCACHED_SUCCESS;

  instance= get_instance(ptr);
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  for (x= 0; x < HASH_LIST_SIZE; x++)
    delete_buckets(ptr, *(instance->list + x));
  /* This strikes me as something that might be unpredictable. */
  /* memset(instance->list, 0, sizeof(bucket *) * HASH_LIST_SIZE); */

  return MEMCACHED_SUCCESS;
}

memcached_return internal_dump(memcached_st *ptr, memcached_dump_func *callback, void *context, uint32_t number_of_callbacks)
{
  memcached_return rc= MEMCACHED_SUCCESS;
  uint32_t x;
  internal_server_state_st *instance;

  instance= get_instance(ptr);
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;


  for (x= 0; x < HASH_LIST_SIZE; x++)
  {
    bucket *bucket_ptr;

    if ((instance->list + x) == NULL)
      continue;

    for (bucket_ptr= *(instance->list + x); bucket_ptr; bucket_ptr= bucket_ptr->next)
    {
      rc= dump_callback_container(ptr, callback, bucket_ptr->key, bucket_ptr->key_length, context, number_of_callbacks);

      if (rc != MEMCACHED_SUCCESS)
        goto end;
    }
  }

end:
  return rc;
}

memcached_return internal_increment(memcached_st *ptr,
                                    const char *key, size_t key_length,
                                    uint32_t offset,
                                    uint64_t *value)
{
  bucket *bucket_ptr;
  char buffer[1024]; /* TODO -> find actual largest size */
  uint64_t new_num;
  int new_length;
  internal_server_state_st *instance;

  instance= get_instance(ptr);
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  bucket_ptr= find_bucket(instance, key, key_length);

  if (bucket_ptr == NULL || bucket_ptr->value_length == 0)
  {
    if (ptr->flags & MEM_NOREPLY)
      return MEMCACHED_SUCCESS;
    return MEMCACHED_NOTFOUND;
  }

  if (bucket_ptr->value_length == 0)
  {
    new_num= offset ? offset : 1;
  }
  else
  {
    new_num= strtoul(bucket_ptr->value, (char **)NULL, 10);
    new_num+= offset ? offset : 1;
  }

  new_length= snprintf(buffer, sizeof(buffer), "%" PRIu64, new_num);

  *value= new_num;

  /* With the preallocated buckets, I don't think we need to do this. */
  /*
  {
    char *new_ptr;

    new_ptr= ptr->call_realloc(ptr, bucket_ptr->value, new_length + 1); /c add a byte for NULL pointer c/
    if (new_ptr == NULL)
      return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

    bucket_ptr->value= new_ptr;
    
  }
  */

  bucket_ptr->value_length= new_length + 1;

  /*
    For now we add in the null byte so that we can eventually fix the code
    to allow hash values be used without the cost of malloc.
  */
  memcpy(bucket_ptr->value, buffer, new_length+1);

  return MEMCACHED_SUCCESS;
}

memcached_return internal_decrement(memcached_st *ptr,
                                    const char *key, size_t key_length,
                                    uint32_t offset,
                                    uint64_t *value)
{
  bucket *bucket_ptr;
  char buffer[1024]; /* TODO -> find actual largest size */
  uint64_t new_num;
  int new_length;
  internal_server_state_st *instance;

  instance= get_instance(ptr);
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  bucket_ptr= find_bucket(instance, key, key_length);

  if (bucket_ptr == NULL || bucket_ptr->value_length == 0)
  {
    if (ptr->flags & MEM_NOREPLY)
      return MEMCACHED_SUCCESS;
    return MEMCACHED_NOTFOUND;
  }

  if (bucket_ptr->value_length == 0)
  {
    new_num= offset ? offset : 1;
    new_num= 0;
  }
  else
  {
    new_num= strtoul(bucket_ptr->value, (char **)NULL, 10);
    new_num-= offset ? offset : 1;
  }

  new_length= snprintf(buffer, sizeof(buffer), "%" PRIu64, new_num);

  *value= new_num;

  /* Same as above w/ preallocated buckets. */
  /* 
  {
    char *new_ptr;

    new_ptr= ptr->call_realloc(ptr, bucket_ptr->value, new_length + 1); /c add a byte for NULL pointer c/
    if (new_ptr == NULL)
      return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

    bucket_ptr->value= new_ptr;
  }
  */

  bucket_ptr->value_length= new_length + 1;

  /*
    For now we add in the null byte so that we can eventually fix the code
    to allow hash values be used without the cost of malloc.
  */
  memcpy(bucket_ptr->value, buffer, new_length+1);

  return MEMCACHED_SUCCESS;
}

memcached_return memcached_internal_clone_hook(memcached_st *org,
                                               memcached_st *clone)
{
   (void)clone;
  internal_server_state_st *instance= get_instance(org);
  if (instance == NULL)
    return MEMCACHED_MEMORY_ALLOCATION_FAILURE;

  ++instance->refcount;
  return MEMCACHED_SUCCESS;
}

void memcached_internal_free_hook(memcached_st *memc)
{
   internal_server_state_st *embedded= memc->hosts[0].embedded_server;
   if (embedded == NULL)
     return ;

   if (embedded->refcount == 0) {
      /* This is the last reference to the cache, nuke it! */
      for (int x= 0; x < HASH_LIST_SIZE; x++)
      {
         bucket *ptr= *(embedded->list + x);

         while (ptr) {
            bucket *tmp= ptr;
            ptr= ptr->next;

            memc->call_free(memc, tmp->value);
            memc->call_free(memc, tmp);
         }
      }
      memc->call_free(memc, embedded);
      memc->hosts[0].embedded_server= NULL;
   } else {
      --embedded->refcount;
   }
}

