#include "common.h"

static struct memcached_protocol_hooks ascii = {
   .storage= memcached_ascii_storage,
   .flush= memcached_ascii_flush,
   .quit_server= memcached_ascii_quit_server,
   .arithmetic= memcached_ascii_arithmetic,
   .arithmetic_with_initial= memcached_ascii_arithmetic_with_initial,
   .delete= memcached_ascii_delete,
   .fetch_stats= memcached_ascii_stats_fetch,
   .version= memcached_ascii_version,
   .mget= memcached_ascii_mget,
   .dump= memcached_ascii_dump,
   .clone_hook= memcached_noop_clone_hook,
   .free_hook= memcached_noop_free_hook,
};
static struct memcached_protocol_hooks binary = {
   .storage= memcached_binary_storage,
   .flush= memcached_binary_flush,
   .quit_server= memcached_binary_quit_server,
   .arithmetic= memcached_binary_arithmetic,
   .arithmetic_with_initial= memcached_binary_arithmetic_with_initial,
   .delete= memcached_binary_delete,
   .fetch_stats= memcached_binary_stats_fetch,
   .version= memcached_binary_version,
   .mget= memcached_binary_mget,
   .dump= memcached_binary_dump,
   .clone_hook= memcached_noop_clone_hook,
   .free_hook= memcached_noop_free_hook,
};

static struct memcached_protocol_hooks embedded = {
   .storage= memcached_internal_storage,
   .flush= memcached_internal_flush,
   .quit_server= memcached_internal_quit_server,
   .arithmetic= memcached_internal_arithmetic,
   .arithmetic_with_initial= memcached_internal_arithmetic_with_initial,
   .delete= memcached_internal_delete,
   .fetch_stats= NULL, /* not supported */
   .version= memcached_internal_version,
   .mget= memcached_internal_mget,
   .dump= memcached_internal_dump,
   .clone_hook= memcached_internal_clone_hook,
   .free_hook= memcached_internal_free_hook,
};

struct memcached_protocol_hooks* get_protocol_hooks(libmemcached_protocol p)
{
   switch (p) {
   case LIBMEMCACHED_ASCII_PROTOCOL:
      return &ascii;
   case LIBMEMCACHED_BINARY_PROTOCOL:
      return &binary;
   case LIBMEMCACHED_EMBEDDED_PROTOCOL:
      return &embedded;
   default:
      /*
      ** This is a private interface, so getting here means that you must
      ** be developing new extensions. Let's abort so that you know you
      ** need to implement this function;
      */
      abort();
   }
}

static inline uint8_t get_com_code(memcached_storage_action verb, bool noreply)
{
  /* 0 isn't a value we want, but GCC 4.2 seems to think ret can otherwise
   * be used uninitialized in this function. FAIL */
  uint8_t ret= 0;

  if (noreply)
    switch (verb)
    {
    case SET_OP:
      ret=PROTOCOL_BINARY_CMD_SETQ;
      break;
    case ADD_OP:
      ret=PROTOCOL_BINARY_CMD_ADDQ;
      break;
    case CAS_OP: /* FALLTHROUGH */
    case REPLACE_OP:
      ret=PROTOCOL_BINARY_CMD_REPLACEQ;
      break;
    case APPEND_OP:
      ret=PROTOCOL_BINARY_CMD_APPENDQ;
      break;
    case PREPEND_OP:
      ret=PROTOCOL_BINARY_CMD_PREPENDQ;
      break;
    default:
      WATCHPOINT_ASSERT(verb);
      break;
    }
  else
    switch (verb)
    {
    case SET_OP:
      ret=PROTOCOL_BINARY_CMD_SET;
      break;
    case ADD_OP:
      ret=PROTOCOL_BINARY_CMD_ADD;
      break;
    case CAS_OP: /* FALLTHROUGH */
    case REPLACE_OP:
      ret=PROTOCOL_BINARY_CMD_REPLACE;
      break;
    case APPEND_OP:
      ret=PROTOCOL_BINARY_CMD_APPEND;
      break;
    case PREPEND_OP:
      ret=PROTOCOL_BINARY_CMD_PREPEND;
      break;
    default:
      WATCHPOINT_ASSERT(verb);
      break;
    }

   return ret;
}

